/*
 * Cryptographic API.
 *
 * Support for Infineon DEU hardware crypto engine.
 *
 * Copyright (c) 2005  Johannes Doering <info@com-style.de>, INFINEON
 *
 * weak key check routine taken from crypto/des.c
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * ---------------------------------------------------------------------------
 * Copyright (c) 2002, Dr Brian Gladman <brg@gladman.me.uk>, Worcester, UK.
 * All rights reserved.
 *
 * LICENSE TERMS
 *
 * The free distribution and use of this software in both source and binary
 * form is allowed (with or without changes) provided that:
 *
 *   1. distributions of this source code include the above copyright
 *      notice, this list of conditions and the following disclaimer;
 *
 *   2. distributions in binary form include the above copyright
 *      notice, this list of conditions and the following disclaimer
 *      in the documentation and/or other associated materials;
 *
 *   3. the copyright holder's name is not used to endorse products
 *      built using this software without specific written permission.
 *
 * ALTERNATIVELY, provided that this notice is retained in full, this product
 * may be distributed under the terms of the GNU General Public License (GPL),
 * in which case the provisions of the GPL apply INSTEAD OF those given above.
 *
 * DISCLAIMER
 *
 * This software is provided 'as is' with no explicit or implied warranties
 * in respect of its properties, including, but not limited to, correctness
 * and/or fitness for purpose.
 * ---------------------------------------------------------------------------
 ** HISTORY
 ** $Date        $Author        $Comment
 *  9th Nov 2007 Teh Kok How    Bug fixes
 * ---------------------------------------------------------------------------
 */

/**
  \addtogroup AMAZON_S_DEU
  \ingroup AMAZON_S_BSP
  \brief amazon_s deu driver module
*/

/*!
  \file amazon_s_deu_des.c
  \ingroup AMAZON_S_DEU
  \brief des driver file
*/

/*!
  \addtogroup AMAZON_S_DEU_FUNCTIONS
  \ingroup AMAZON_S_DEU
  \brief amazon_s deu driver functions
*/

#include <linux/version.h>
#include <linux/module.h>
#include <linux/init.h>
#include <linux/types.h>
#include <linux/errno.h>
#include <linux/crypto.h>
#include <crypto/algapi.h>
#include <linux/interrupt.h>
#include <asm/byteorder.h>
#include <linux/delay.h>
#include <amazon_s_deu.h>

#if 0
     #define CRTCL_SECT_INIT
     #define CRTCL_SECT_START       local_irq_save(flag)
     #define CRTCL_SECT_END         local_irq_restore(flag)
#else
     static spinlock_t lock;
     #define CRTCL_SECT_INIT        spin_lock_init(&lock)
     #define CRTCL_SECT_START       spin_lock_irqsave(&lock, flag)
     #define CRTCL_SECT_END         spin_unlock_irqrestore(&lock, flag)
#endif

//#define CRYPTO_DEBUG
#ifdef CRYPTO_DEBUG
extern char debug_level;
#define DPRINTF(level, format, args...) if (level < debug_level) printk(KERN_INFO "[%s %s %d]: " format, __FILE__, __func__, __LINE__, ##args);
#else
#define DPRINTF(level, format, args...)
#endif

#ifdef CONFIG_CRYPTO_DEV_AMAZON_S_DES
#include <amazon_s.h>
#include <amazon_s_deu_structs.h>
#define DES_3DES_START  AMAZON_S_DES_CON
#endif
#ifdef CONFIG_CRYPTO_DEV_AMAZON_S_DMA
#include <amazon_s_deu_dma.h>
#include <amazon_s_dma.h>
#include <irq.h>
#endif

#define DES_3DES_START  AMAZON_S_DES_CON

#define DES_KEY_SIZE            8
#define DES_EXPKEY_WORDS        32
#define DES_BLOCK_SIZE          8

#define DES3_EDE_KEY_SIZE       (3 * DES_KEY_SIZE)
#define DES3_EDE_EXPKEY_WORDS   (3 * DES_EXPKEY_WORDS)
#define DES3_EDE_BLOCK_SIZE     DES_BLOCK_SIZE

struct des_ctx {
#if 0 // share des_ctx with des3_ede_ctx
        int controlr_M;
        int key_length;
        u8 iv[DES_BLOCK_SIZE];
        u32 expkey[DES_EXPKEY_WORDS];
};
struct des3_ede_ctx {
#endif // share des_ctx with des3_ede_ctx
        int controlr_M;
        int key_length;
        u8 iv[DES_BLOCK_SIZE];
        u32 expkey[DES3_EDE_EXPKEY_WORDS];
};

extern int disable_multiblock;
extern int disable_deudma;

extern int deu_dma_release (void);

/*! \fn static u32 endian_swap(u32 input)
    \ingroup AMAZON_S_DEU_FUNCTIONS
    \brief perform dword level endian swap
    \param input value of dword that requires to be swapped
*/
static u32 endian_swap(u32 input)
{
#ifdef CONFIG_CRYPTO_DEV_AMAZON_S_DMA
    u8 *ptr = (u8 *)&input;

    return ((ptr[3] << 24) | (ptr[2] << 16) | (ptr[1] << 8) | ptr[0]);
#else
    return input;
#endif
}

/*! \fn static void des_chip_init (void)
    \ingroup AMAZON_S_DEU_FUNCTIONS
    \brief initialize DES hardware
*/
static void des_chip_init (void)
{
        volatile struct des_t *des = (struct des_t *) DES_3DES_START;

#ifndef CONFIG_CRYPTO_DEV_AMAZON_S_DMA
        // start crypto engine with write to ILR
        des->controlr.SM = 1;
        des->controlr.ARS = 1;
#else
        des->controlr.SM = 1;
        des->controlr.ARS = 0; // 0 for dma

        // needed for dma mode
        des->controlr.NDC = 1;
        asm("sync");
        des->controlr.ENDI = 0;
#endif
}

/*! \fn  static int des_setkey(struct crypto_tfm *tfm, const u8 *key, unsigned int keylen)
    \ingroup AMAZON_S_DEU_FUNCTIONS
    \brief sets DES key
    \param tfm linux crypto algo transform
    \param key input key
    \param keylen key length
*/
static int des_setkey(struct crypto_tfm *tfm, const u8 *key,
                      unsigned int keylen)
{
        struct des_ctx *dctx = crypto_tfm_ctx(tfm);

        printk("setkey in %s\n", __FILE__);

        dctx->controlr_M = 0;   // des
        dctx->key_length = keylen;

        memcpy ((u8 *) (dctx->expkey), key, keylen);

        return 0;
}

/*! \fn  static void amazon_s_deu_des(void *ctx_arg, u8 *out_arg, const u8 *in_arg, u8 *iv_arg, u32 nbytes, int encdec, int mode)
    \ingroup AMAZON_S_DEU_FUNCTIONS
    \brief main interface to DES hardware
    \param ctx_arg crypto algo context
    \param out_arg output bytestream
    \param in_arg input bytestream
    \param iv_arg initialization vector
    \param nbytes length of bytestream
    \param encdec 1 for encrypt; 0 for decrypt
    \param mode operation mode such as ebc, cbc
*/
#ifndef CONFIG_CRYPTO_DEV_AMAZON_S_DMA
static void amazon_s_deu_des (void *ctx_arg, u8 *out_arg, const u8 *in_arg,
             u8 *iv_arg, u32 nbytes, int encdec, int mode)
#else
static void amazon_s_deu_des_core (void *ctx_arg, u8 *out_arg, const u8 *in_arg,
             u8 *iv_arg, u32 nbytes, int encdec, int mode)
#endif
{
        volatile struct des_t *des = (struct des_t *) DES_3DES_START;
        struct des_ctx *dctx = ctx_arg;
        u32 *key = dctx->expkey;
        unsigned long flag;

        CRTCL_SECT_START;

        des->controlr.M = dctx->controlr_M;
        if (dctx->controlr_M == 0)      // des
        {
                des->K1HR = endian_swap(*((u32 *) key + 0));
                des->K1LR = endian_swap(*((u32 *) key + 1));
                //printk("*((u32 *) key + 0) = %08x\n", *((u32 *) key + 0));
                //printk("*((u32 *) key + 1) = %08x\n", *((u32 *) key + 1));

        }
        else {
                /* Hardware Section */
                switch (dctx->key_length) {
                case 24:
                        des->K3HR = endian_swap(*((u32 *) key + 4));
                        des->K3LR = endian_swap(*((u32 *) key + 5));
                        /* no break; */

                case 16:
                        des->K2HR = endian_swap(*((u32 *) key + 2));
                        des->K2LR = endian_swap(*((u32 *) key + 3));

                        /* no break; */
                case 8:
                        des->K1HR = endian_swap(*((u32 *) key + 0));
                        des->K1LR = endian_swap(*((u32 *) key + 1));
                        break;

                default:
                        CRTCL_SECT_END;
                        return;
                }
        }

        //hexdump(in_arg, nbytes);

        //printk ("encdec %x\n", encdec);
        //printk ("mode %x\n", mode);
        //printk ("nbytes %u\n", nbytes);

        des->controlr.E_D = !encdec;    //encryption
        des->controlr.O = mode; //0 ECB 1 CBC 2 OFB 3 CFB 4 CTR hexdump(prin,sizeof(*des));
        if (mode > 0) {
                des->IVHR = endian_swap(*(u32 *) iv_arg);
                des->IVLR = endian_swap(*((u32 *) iv_arg + 1));
                //printk("des->IVHR  = %08x\n", des->IVHR);
                //printk("des->IVLR  = %08x\n", des->IVLR);
        };

        {
#ifndef CONFIG_CRYPTO_DEV_AMAZON_S_DMA
                int i = 0;
                int nblocks = 0;

                nblocks = nbytes / 4;

                for (i = 0; i < nblocks; i += 2) {
                        /* wait for busy bit to clear */

                        /*--- Workaround ----------------------------------------------------
                        do a dummy read to the busy flag because it is not raised early
                        enough in CFB/OFB 3DES modes */
 #ifdef CRYPTO_DEBUG
                        printk ("ihr: %x\n", (*((u32 *) in_arg + i)));
                        printk ("ilr: %x\n", (*((u32 *) in_arg + 1 + i)));
 #endif
                        des->IHR = (*((u32 *) in_arg + i));
                        des->ILR = (*((u32 *) in_arg + 1 + i)); /* start crypto */
                        //printk("des->IHR  = %08x\n", des->IHR);
                        //printk("des->ILR  = %08x\n", des->ILR);
                        //printk("in_arg  = %08x\n", (*((u32 *) in_arg + i)));
                        //printk("in_arg2  = %08x\n", (*((u32 *) in_arg + 1 + i)));

                        while (des->controlr.BUS) {
                                // this will not take long
                                //printk("des->controlr.BUS\n");
                        }

                        *((u32 *) out_arg + 0 + i) = des->OHR;
                        *((u32 *) out_arg + 1 + i) = des->OLR;
                        //printk("des->OHR  = %08x\n", des->OHR);
                        //printk("des->OLR  = %08x\n", des->OLR);

                }

#else // dma

                //volatile struct deu_dma_t *dma = (struct deu_dma_t *) DMA_CON;
                struct dma_device_info *dma_device = ifx_deu[0].dma_device;
                _ifx_deu_device *pDev = ifx_deu;

                int wlen = 0;
                u32 timeout = 0;
                u8 *out_dma = NULL;

                pDev->len = nbytes;
                pDev->packet_size = nbytes;
                pDev->src = (u8 *)in_arg;

                pDev->dst = out_arg;
                pDev->dst_count = 0;

                //dma->controlr.ALGO = 0;       //DES
                *AMAZON_S_CON = 0x00000201; //burst size=4 words; algo=DES; enable dma

                wlen = dma_device_write (dma_device, (u8 *)in_arg, nbytes, NULL);

                if (wlen != nbytes) {
                        CRTCL_SECT_END;
                        printk (KERN_ERR "[%s %s %d]: dma_device_write fail!\n", __FILE__, __func__, __LINE__);
                        return; // -EINVAL;
                }

 #if 0
                {
  #define CNT_TEST	10
                        volatile int cnt;

                        for (cnt = 0; cnt < CNT_TEST; cnt++)
                            udelay(10);
                }
 #endif

                while(((*AMAZON_S_CON & 0x00000080) == 0x00000080)){
                        // this will not take long
                };

                while (des->controlr.BUS) {
                        // this will not take long
                }

                // polling DMA rx channel
                while ((pDev->recv_count =
                        dma_device_read (dma_device, &out_dma, NULL)) == 0) {
                        timeout++;

                        if (timeout >= 333000) {
                                deu_dma_release ();
                                CRTCL_SECT_END;
                                printk (KERN_ERR "[%s %s %d]: timeout!!\n", __FILE__, __func__, __LINE__);
                                return; // -EINVAL;
                        }
                }

 #if 0
                for (cnt = 0; cnt < CNT_TEST; cnt++)
                    udelay(10);
 #endif

                while(((*AMAZON_S_CON & 0x00000080) == 0x00000080)){
                        // this will not take long
                };

                while (des->controlr.BUS) {
                        // this will not take long
                }

                memcpy (out_arg, out_dma, nbytes);
#endif // dma
        }

        //tc.chen: copy iv_arg back
        if (mode > 0) {
                *(u32 *) iv_arg = des->IVHR;
                *((u32 *) iv_arg + 1) = des->IVLR;
                *(u32 *) iv_arg = endian_swap(*(u32 *) iv_arg);
                *((u32 *) iv_arg + 1) = endian_swap(*((u32 *) iv_arg + 1));
        };

        CRTCL_SECT_END;
}

#ifdef CONFIG_CRYPTO_DEV_AMAZON_S_DMA
static void amazon_s_deu_des (void *ctx_arg, u8 *out_arg, const u8 *in_arg,
             u8 *iv_arg, u32 nbytes, int encdec, int mode)
{
	u32 remain = nbytes;
	u32 inc;

        while (remain > 0)
        {
            if (remain >= DEU_MAX_PACKET_SIZE)
            {
                inc = DEU_MAX_PACKET_SIZE;
            }
            else
            {
                inc = remain;
            }

            remain -= inc;

            amazon_s_deu_des_core(ctx_arg, out_arg, in_arg, iv_arg, inc, encdec, mode);

            out_arg += inc;
            in_arg += inc;
        }

}
#endif


//definitions from linux/include/crypto.h:
//#define CRYPTO_TFM_MODE_ECB           0x00000001
//#define CRYPTO_TFM_MODE_CBC           0x00000002
//#define CRYPTO_TFM_MODE_CFB           0x00000004
//#define CRYPTO_TFM_MODE_CTR           0x00000008
//#define CRYPTO_TFM_MODE_OFB           0x00000010 // not even defined
//but hardware definition: 0 ECB 1 CBC 2 OFB 3 CFB 4 CTR

/*! \fn static void amazon_s_deu_des_ecb (void *ctx, uint8_t *dst, const uint8_t *src, uint8_t *iv, size_t nbytes, int encdec, int inplace)
    \ingroup AMAZON_S_DEU_FUNCTIONS
    \brief sets DES hardware to ECB mode
    \param ctx crypto algo context
    \param dst output bytestream
    \param src input bytestream
    \param iv initialization vector
    \param nbytes length of bytestream
    \param encdec 1 for encrypt; 0 for decrypt
    \param inplace not used
*/
static void amazon_s_deu_des_ecb (void *ctx, uint8_t *dst, const uint8_t *src,
                uint8_t *iv, size_t nbytes, int encdec, int inplace)
{
        amazon_s_deu_des (ctx, dst, src, NULL, nbytes, encdec, 0);
}

/*! \fn static void amazon_s_deu_des_cbc (void *ctx, uint8_t *dst, const uint8_t *src, uint8_t *iv, size_t nbytes, int encdec, int inplace)
    \ingroup AMAZON_S_DEU_FUNCTIONS
    \brief sets DES hardware to CBC mode
    \param ctx crypto algo context
    \param dst output bytestream
    \param src input bytestream
    \param iv initialization vector
    \param nbytes length of bytestream
    \param encdec 1 for encrypt; 0 for decrypt
    \param inplace not used
*/
static void amazon_s_deu_des_cbc (void *ctx, uint8_t *dst, const uint8_t *src,
                uint8_t *iv, size_t nbytes, int encdec, int inplace)
{
        amazon_s_deu_des (ctx, dst, src, iv, nbytes, encdec, 1);
}

#if (LINUX_VERSION_CODE < KERNEL_VERSION(2,6,20))

/*! \fn static void amazon_s_deu_des_ofb (void *ctx, uint8_t *dst, const uint8_t *src, uint8_t *iv, size_t nbytes, int encdec, int inplace)
    \ingroup AMAZON_S_DEU_FUNCTIONS
    \brief sets DES hardware to OFB mode
    \param ctx crypto algo context
    \param dst output bytestream
    \param src input bytestream
    \param iv initialization vector
    \param nbytes length of bytestream
    \param encdec 1 for encrypt; 0 for decrypt
    \param inplace not used
*/
static void amazon_s_deu_des_ofb (void *ctx, uint8_t *dst, const uint8_t *src,
                uint8_t *iv, size_t nbytes, int encdec, int inplace)
{
        amazon_s_deu_des (ctx, dst, src, iv, nbytes, encdec, 2);
}

/*! \fn static void amazon_s_deu_des_cfb (void *ctx, uint8_t *dst, const uint8_t *src, uint8_t *iv, size_t nbytes, int encdec, int inplace)
    \ingroup AMAZON_S_DEU_FUNCTIONS
    \brief sets DES hardware to CFB mode
    \param ctx crypto algo context
    \param dst output bytestream
    \param src input bytestream
    \param iv initialization vector
    \param nbytes length of bytestream
    \param encdec 1 for encrypt; 0 for decrypt
    \param inplace not used
*/
static void amazon_s_deu_des_cfb (void *ctx, uint8_t *dst, const uint8_t *src,
                uint8_t *iv, size_t nbytes, int encdec, int inplace)
{
        amazon_s_deu_des (ctx, dst, src, iv, nbytes, encdec, 3);
}

/*! \fn static void amazon_s_deu_des_ctr (void *ctx, uint8_t *dst, const uint8_t *src, uint8_t *iv, size_t nbytes, int encdec, int inplace)
    \ingroup AMAZON_S_DEU_FUNCTIONS
    \brief sets DES hardware to CTR mode
    \param ctx crypto algo context
    \param dst output bytestream
    \param src input bytestream
    \param iv initialization vector
    \param nbytes length of bytestream
    \param encdec 1 for encrypt; 0 for decrypt
    \param inplace not used
*/
static void amazon_s_deu_des_ctr (void *ctx, uint8_t *dst, const uint8_t *src,
                uint8_t *iv, size_t nbytes, int encdec, int inplace)
{
        amazon_s_deu_des (ctx, dst, src, iv, nbytes, encdec, 4);
}

#endif  //      (LINUX_VERSION_CODE < KERNEL_VERSION(2,6,20))

/*! \fn static void des_encrypt (struct crypto_tfm *tfm, uint8_t *out, const uint8_t *in)
    \ingroup AMAZON_S_DEU_FUNCTIONS
    \brief encrypt DES_BLOCK_SIZE of data
    \param tfm linux crypto algo transform
    \param out output bytestream
    \param in input bytestream
*/
static void des_encrypt (struct crypto_tfm *tfm, uint8_t * out, const uint8_t * in)
{
        struct des_ctx *ctx = crypto_tfm_ctx(tfm);
        amazon_s_deu_des (ctx, out, in, NULL, DES_BLOCK_SIZE,
                    CRYPTO_DIR_ENCRYPT, 0);

}

/*! \fn static void des_decrypt (struct crypto_tfm *tfm, uint8_t *out, const uint8_t *in)
    \ingroup AMAZON_S_DEU_FUNCTIONS
    \brief encrypt DES_BLOCK_SIZE of data
    \param tfm linux crypto algo transform
    \param out output bytestream
    \param in input bytestream
*/
static void des_decrypt (struct crypto_tfm *tfm, uint8_t * out, const uint8_t * in)
{
        struct des_ctx *ctx = crypto_tfm_ctx(tfm);
        amazon_s_deu_des (ctx, out, in, NULL, DES_BLOCK_SIZE,
                    CRYPTO_DIR_DECRYPT, 0);
}

/*
 * RFC2451:
 *
 *   For DES-EDE3, there is no known need to reject weak or
 *   complementation keys.  Any weakness is obviated by the use of
 *   multiple keys.
 *
 *   However, if the first two or last two independent 64-bit keys are
 *   equal (k1 == k2 or k2 == k3), then the DES3 operation is simply the
 *   same as DES.  Implementers MUST reject keys that exhibit this
 *   property.
 *
 */

/*! \fn  static int des3_ede_setkey(struct crypto_tfm *tfm, const u8 *key, unsigned int keylen)
    \ingroup AMAZON_S_DEU_FUNCTIONS
    \brief sets 3DES key
    \param tfm linux crypto algo transform
    \param key input key
    \param keylen key length
*/
static int des3_ede_setkey(struct crypto_tfm *tfm, const u8 *key,
                           unsigned int keylen)
{
        struct des_ctx *dctx = crypto_tfm_ctx(tfm);

        printk("setkey in %s\n", __FILE__);

        dctx->controlr_M = keylen / 8 + 1;      // 3DES EDE1 / EDE2 / EDE3 Mode
        dctx->key_length = keylen;

        memcpy ((u8 *) (dctx->expkey), key, keylen);

        return 0;
}

static struct crypto_alg ifxdeu_des_alg = {
        .cra_name               =       "des",
        .cra_driver_name        =       "ifxdeu-des",
        .cra_flags              =       CRYPTO_ALG_TYPE_CIPHER,
        .cra_blocksize          =       DES_BLOCK_SIZE,
        .cra_ctxsize            =       sizeof(struct des_ctx),
        .cra_module             =       THIS_MODULE,
        .cra_alignmask          =       3,
        .cra_list               =       LIST_HEAD_INIT(ifxdeu_des_alg.cra_list),
        .cra_u                  =       { .cipher = {
        .cia_min_keysize        =       DES_KEY_SIZE,
        .cia_max_keysize        =       DES_KEY_SIZE,
        .cia_setkey             =       des_setkey,
        .cia_encrypt            =       des_encrypt,
        .cia_decrypt            =       des_decrypt } }
};

static struct crypto_alg ifxdeu_des3_ede_alg = {
        .cra_name               =       "des3_ede",
        .cra_driver_name        =       "ifxdeu-des3_ede",
        .cra_flags              =       CRYPTO_ALG_TYPE_CIPHER,
        .cra_blocksize          =       DES_BLOCK_SIZE,
        .cra_ctxsize            =       sizeof(struct des_ctx),
        .cra_module             =       THIS_MODULE,
        .cra_alignmask          =       3,
        .cra_list               =       LIST_HEAD_INIT(ifxdeu_des3_ede_alg.cra_list),
        .cra_u                  =       { .cipher = {
        .cia_min_keysize        =       DES_KEY_SIZE,
        .cia_max_keysize        =       DES_KEY_SIZE,
        .cia_setkey             =       des3_ede_setkey,
        .cia_encrypt            =       des_encrypt,
        .cia_decrypt            =       des_decrypt } }
};

/*! \fn  static int ecb_des_encrypt(struct blkcipher_desc *desc, struct scatterlist *dst, struct scatterlist *src, unsigned int nbytes)
    \ingroup AMAZON_S_DEU_FUNCTIONS
    \brief ECB DES encrypt using linux crypto blkcipher
    \param desc blkcipher descriptor
    \param dst output scatterlist
    \param src input scatterlist
    \param nbytes data size in bytes
*/
static int ecb_des_encrypt(struct blkcipher_desc *desc,
                           struct scatterlist *dst, struct scatterlist *src,
                           unsigned int nbytes)
{
        struct des_ctx *ctx = crypto_blkcipher_ctx(desc->tfm);
        struct blkcipher_walk walk;
        int err;

        blkcipher_walk_init(&walk, dst, src, nbytes);
        err = blkcipher_walk_virt(desc, &walk);

        while ((nbytes = walk.nbytes)) {
                nbytes -= (nbytes % DES_BLOCK_SIZE);
                amazon_s_deu_des_ecb(ctx, walk.dst.virt.addr, walk.src.virt.addr,
                               NULL, nbytes, CRYPTO_DIR_ENCRYPT, 0);
                nbytes &= DES_BLOCK_SIZE - 1;
                err = blkcipher_walk_done(desc, &walk, nbytes);
        }

        return err;
}

/*! \fn  static int ecb_des_decrypt(struct blkcipher_desc *desc, struct scatterlist *dst, struct scatterlist *src, unsigned int nbytes)
    \ingroup AMAZON_S_DEU_FUNCTIONS
    \brief ECB DES decrypt using linux crypto blkcipher
    \param desc blkcipher descriptor
    \param dst output scatterlist
    \param src input scatterlist
    \param nbytes data size in bytes
*/
static int ecb_des_decrypt(struct blkcipher_desc *desc,
                           struct scatterlist *dst, struct scatterlist *src,
                           unsigned int nbytes)
{
        struct des_ctx *ctx = crypto_blkcipher_ctx(desc->tfm);
        struct blkcipher_walk walk;
        int err;

        DPRINTF(1, "\n");
        blkcipher_walk_init(&walk, dst, src, nbytes);
        err = blkcipher_walk_virt(desc, &walk);

        while ((nbytes = walk.nbytes)) {
                nbytes -= (nbytes % DES_BLOCK_SIZE);
                amazon_s_deu_des_ecb(ctx, walk.dst.virt.addr, walk.src.virt.addr,
                               NULL, nbytes, CRYPTO_DIR_DECRYPT, 0);
                nbytes &= DES_BLOCK_SIZE - 1;
                err = blkcipher_walk_done(desc, &walk, nbytes);
        }

        return err;
}

static struct crypto_alg ifxdeu_ecb_des_alg = {
        .cra_name               =       "ecb(des)",
        .cra_driver_name        =       "ifxdeu-ecb(des)",
        .cra_flags              =       CRYPTO_ALG_TYPE_BLKCIPHER,
        .cra_blocksize          =       DES_BLOCK_SIZE,
        .cra_ctxsize            =       sizeof(struct des_ctx),
        .cra_type               =       &crypto_blkcipher_type,
        .cra_module             =       THIS_MODULE,
        .cra_list               =       LIST_HEAD_INIT(ifxdeu_ecb_des_alg.cra_list),
        .cra_u                  =       {
                .blkcipher = {
                        .min_keysize            =       DES_KEY_SIZE,
                        .max_keysize            =       DES_KEY_SIZE,
                        .setkey                 =       des_setkey,
                        .encrypt                =       ecb_des_encrypt,
                        .decrypt                =       ecb_des_decrypt,
                }
        }
};

static struct crypto_alg ifxdeu_ecb_des3_ede_alg = {
        .cra_name               =       "ecb(des3_ede)",
        .cra_driver_name        =       "ifxdeu-ecb(des3_ede)",
        .cra_flags              =       CRYPTO_ALG_TYPE_BLKCIPHER,
        .cra_blocksize          =       DES3_EDE_BLOCK_SIZE,
        .cra_ctxsize            =       sizeof(struct des_ctx),
        .cra_type               =       &crypto_blkcipher_type,
        .cra_module             =       THIS_MODULE,
        .cra_list               =       LIST_HEAD_INIT(ifxdeu_ecb_des3_ede_alg.cra_list),
        .cra_u                  =       {
                .blkcipher = {
                        .min_keysize            =       DES3_EDE_KEY_SIZE,
                        .max_keysize            =       DES3_EDE_KEY_SIZE,
                        .setkey                 =       des3_ede_setkey,
                        .encrypt                =       ecb_des_encrypt,
                        .decrypt                =       ecb_des_decrypt,
                }
        }
};

/*! \fn  static int cbc_des_encrypt(struct blkcipher_desc *desc, struct scatterlist *dst, struct scatterlist *src, unsigned int nbytes)
    \ingroup AMAZON_S_DEU_FUNCTIONS
    \brief CBC DES encrypt using linux crypto blkcipher
    \param desc blkcipher descriptor
    \param dst output scatterlist
    \param src input scatterlist
    \param nbytes data size in bytes
*/
static int cbc_des_encrypt(struct blkcipher_desc *desc,
                           struct scatterlist *dst, struct scatterlist *src,
                           unsigned int nbytes)
{
        struct des_ctx *ctx = crypto_blkcipher_ctx(desc->tfm);
        struct blkcipher_walk walk;
        int err;

        DPRINTF(1, "\n");
        blkcipher_walk_init(&walk, dst, src, nbytes);
        err = blkcipher_walk_virt(desc, &walk);

        while ((nbytes = walk.nbytes)) {
                u8 *iv = walk.iv;
                //printk("iv = %08x\n", *(u32 *)iv);
                nbytes -= (nbytes % DES_BLOCK_SIZE);
                amazon_s_deu_des_cbc(ctx, walk.dst.virt.addr, walk.src.virt.addr,
                               iv, nbytes, CRYPTO_DIR_ENCRYPT, 0);
                nbytes &= DES_BLOCK_SIZE - 1;
                err = blkcipher_walk_done(desc, &walk, nbytes);
        }

        return err;
}

/*! \fn  static int cbc_des_decrypt(struct blkcipher_desc *desc, struct scatterlist *dst, struct scatterlist *src, unsigned int nbytes)
    \ingroup AMAZON_S_DEU_FUNCTIONS
    \brief CBC DES decrypt using linux crypto blkcipher
    \param desc blkcipher descriptor
    \param dst output scatterlist
    \param src input scatterlist
    \param nbytes data size in bytes
*/
static int cbc_des_decrypt(struct blkcipher_desc *desc,
                           struct scatterlist *dst, struct scatterlist *src,
                           unsigned int nbytes)
{
        struct des_ctx *ctx = crypto_blkcipher_ctx(desc->tfm);
        struct blkcipher_walk walk;
        int err;

        DPRINTF(1, "\n");
        blkcipher_walk_init(&walk, dst, src, nbytes);
        err = blkcipher_walk_virt(desc, &walk);

        while ((nbytes = walk.nbytes)) {
                u8 *iv = walk.iv;
                //printk("iv = %08x\n", *(u32 *)iv);
                nbytes -= (nbytes % DES_BLOCK_SIZE);
                amazon_s_deu_des_cbc(ctx, walk.dst.virt.addr, walk.src.virt.addr,
                               iv, nbytes, CRYPTO_DIR_DECRYPT, 0);
                nbytes &= DES_BLOCK_SIZE - 1;
                err = blkcipher_walk_done(desc, &walk, nbytes);
        }

        return err;
}

static struct crypto_alg ifxdeu_cbc_des_alg = {
        .cra_name               =       "cbc(des)",
        .cra_driver_name        =       "ifxdeu-cbc(des)",
        .cra_flags              =       CRYPTO_ALG_TYPE_BLKCIPHER,
        .cra_blocksize          =       DES_BLOCK_SIZE,
        .cra_ctxsize            =       sizeof(struct des_ctx),
        .cra_type               =       &crypto_blkcipher_type,
        .cra_module             =       THIS_MODULE,
        .cra_list               =       LIST_HEAD_INIT(ifxdeu_cbc_des_alg.cra_list),
        .cra_u                  =       {
                .blkcipher = {
                        .min_keysize            =       DES_KEY_SIZE,
                        .max_keysize            =       DES_KEY_SIZE,
                        .ivsize                 =       DES_BLOCK_SIZE,
                        .setkey                 =       des_setkey,
                        .encrypt                =       cbc_des_encrypt,
                        .decrypt                =       cbc_des_decrypt,
                }
        }
};

static struct crypto_alg ifxdeu_cbc_des3_ede_alg = {
        .cra_name               =       "cbc(des3_ede)",
        .cra_driver_name        =       "ifxdeu-cbc(des3_ede)",
        .cra_flags              =       CRYPTO_ALG_TYPE_BLKCIPHER,
        .cra_blocksize          =       DES3_EDE_BLOCK_SIZE,
        .cra_ctxsize            =       sizeof(struct des_ctx),
        .cra_type               =       &crypto_blkcipher_type,
        .cra_module             =       THIS_MODULE,
        .cra_list               =       LIST_HEAD_INIT(ifxdeu_cbc_des3_ede_alg.cra_list),
        .cra_u                  =       {
                .blkcipher = {
                        .min_keysize            =       DES3_EDE_KEY_SIZE,
                        .max_keysize            =       DES3_EDE_KEY_SIZE,
                        .ivsize                 =       DES_BLOCK_SIZE,
                        .setkey                 =       des3_ede_setkey,
                        .encrypt                =       cbc_des_encrypt,
                        .decrypt                =       cbc_des_decrypt,
                }
        }
};

/*! \fn  int __init ifxdeu_init_des (void)
    \ingroup AMAZON_S_DEU_FUNCTIONS
    \brief initialize des driver
*/
int __init ifxdeu_init_des (void)
{
        int ret = 0;

#if (LINUX_VERSION_CODE < KERNEL_VERSION(2,6,20))
        if (!disable_multiblock) {
                ifxdeu_des_alg.cra_u.cipher.cia_max_nbytes = DES_BLOCK_SIZE;    //(size_t)-1;
                ifxdeu_des_alg.cra_u.cipher.cia_req_align = 16;
                ifxdeu_des_alg.cra_u.cipher.cia_ecb = amazon_s_deu_des_ecb;
                ifxdeu_des_alg.cra_u.cipher.cia_cbc = amazon_s_deu_des_cbc;
                ifxdeu_des_alg.cra_u.cipher.cia_cfb = amazon_s_deu_des_cfb;
                ifxdeu_des_alg.cra_u.cipher.cia_ofb = amazon_s_deu_des_ofb;
        }
#endif

        ret = crypto_register_alg(&ifxdeu_des_alg);
        if (ret < 0)
                goto des_err;

        ret = crypto_register_alg(&ifxdeu_ecb_des_alg);
        if (ret < 0)
                goto ecb_des_err;

        ret = crypto_register_alg(&ifxdeu_cbc_des_alg);
        if (ret < 0)
                goto cbc_des_err;

        ret = crypto_register_alg(&ifxdeu_des3_ede_alg);
        if (ret < 0)
                goto des3_ede_err;

        ret = crypto_register_alg(&ifxdeu_ecb_des3_ede_alg);
        if (ret < 0)
                goto ecb_des3_ede_err;

        ret = crypto_register_alg(&ifxdeu_cbc_des3_ede_alg);
        if (ret < 0)
                goto cbc_des3_ede_err;

        des_chip_init();

        CRTCL_SECT_INIT;

        printk (KERN_NOTICE "IFX DEU DES initialized%s%s.\n", disable_multiblock ? "" : " (multiblock)", disable_deudma ? "" : " (DMA)");
        return ret;

des_err:
        crypto_unregister_alg(&ifxdeu_des_alg);
        printk(KERN_ERR "IFX des initialization failed!\n");
        return ret;
ecb_des_err:
        crypto_unregister_alg(&ifxdeu_ecb_des_alg);
        printk (KERN_ERR "IFX ecb_des initialization failed!\n");
        return ret;
cbc_des_err:
        crypto_unregister_alg(&ifxdeu_cbc_des_alg);
        printk (KERN_ERR "IFX cbc_des initialization failed!\n");
        return ret;
des3_ede_err:
        crypto_unregister_alg(&ifxdeu_des3_ede_alg);
        printk(KERN_ERR "IFX des3_ede initialization failed!\n");
        return ret;
ecb_des3_ede_err:
        crypto_unregister_alg(&ifxdeu_ecb_des3_ede_alg);
        printk (KERN_ERR "IFX ecb_des3_ede initialization failed!\n");
        return ret;
cbc_des3_ede_err:
        crypto_unregister_alg(&ifxdeu_cbc_des3_ede_alg);
        printk (KERN_ERR "IFX cbc_des3_ede initialization failed!\n");
        return ret;

}

/*! \fn  void __exit ifxdeu_fini_des (void)
    \ingroup AMAZON_S_DEU_FUNCTIONS
    \brief unregister des driver
*/
void __exit ifxdeu_fini_des (void)
{
        crypto_unregister_alg (&ifxdeu_des_alg);
        crypto_unregister_alg (&ifxdeu_ecb_des_alg);
        crypto_unregister_alg (&ifxdeu_cbc_des_alg);
        crypto_unregister_alg (&ifxdeu_des3_ede_alg);
        crypto_unregister_alg (&ifxdeu_ecb_des3_ede_alg);
        crypto_unregister_alg (&ifxdeu_cbc_des3_ede_alg);
}




